(in-package "URI-TEMPLATE")

(defvar *decode-uri-string* t)

(defvar %uri-protocol)
(defvar %uri-host)
(defvar %uri-path)
(defvar %uri-directory)
(defvar %uri-file)
(defvar %uri-query-string)
(defvar %uri-hash)

(defmacro bind-standard-uri-components (uri-reference &body body)
  (let ((t1 (gensym))
        (t2 (gensym)))
    `(destructuring-bind (&optional ,t1 %uri-protocol %uri-host %uri-path %uri-directory ,t2 %uri-file %uri-query-string %uri-hash)
         (coerce (nth-value 1
                            (cl-ppcre:scan-to-strings
                             ;; regex adapted from http://codesnippets.joyent.com/posts/show/523, could use improvement
                             "^((http[s]?|ftp):\\/)?\\/?([^:\\/\\s]+)(((\\/\\w+)*\\/)([\\w\\-\\.]+[^#?\\s]+))\\??([^#]*)?#?([#\\w\\-]+)?$" 
                             ,uri-reference))
                 'list)
       (declare (ignore ,t1 ,t2))
       ,@body)))

(defmacro uri-template-bind ((template) uri &body body)
  "Binds URI template placeholders (which must be either symbols or
lists like (#'parse-fn var), much like arguments to
cl-ppcre:register-groups-bind) in given URI to the specified
variables, as well as binding a set of standard special
variables (%uri-protocol, %uri-host, etc.) to their respective parts
of the given URI."
  (let* ((template (cdr template)) ;; template is expected to look like output of #U: '(uri-template &rest args)
         (template-bindings (mapcar #'second (remove-if #'stringp template)))
         (template-vars (loop for x in template-bindings when (symbolp x) collect x))
         (uri-var (gensym)))
    `(let ((,uri-var ,uri))
       (bind-standard-uri-components ,uri-var
         (register-groups-bind ,template-bindings
             ('(:sequence :start-anchor ,@(substitute-if-not '(:REGISTER (:GREEDY-REPETITION 0 NIL :EVERYTHING))
                                                             #'stringp template)
                :end-anchor)
               ,uri-var)
           ,(when *decode-uri-string*
                  `(setf ,@(mapcan (lambda (var)
                                     `(,var (kmrcl:decode-uri-string ,var)))
                                   template-vars)))
           ,@body)))))
